/* ------------------------------------------------------------------
 * 
 *  main.c   Simple Servo controller
 *
 *  This program uses a high priority task and a hardware timer to handle the servo pulses
 *  Servo values are updated by pushing desired values onto a queue
 *  
 *  This method burns up to 2ms every 10ms to handle the servo pulses. If you need more 
 *  CPU time for processing, you can lengthen the time between servo pulses or 
 *  rewrite the code using interrupts.
 * 
 *  Mason, Seattle U, 2013
 * 
 *  See freeRTOS readme
 * 
 * --------------------------------------------------------------- */



#include "MX4cKsetup.h"
#include "MX4cKsettings.h"


// the RTOS is configured with a 1ms tick length
#include "RTOS/FreeRTOS.h"
#include "RTOS/task.h"
#include "RTOS/queue.h"
#include "RTOS/timers.h"
#include "RTOS/semphr.h"

#define UP           0
#define DOWN         1
#define SERVOUPDATE 10      // milliseconds between servo updates

// forward declaration
static void Servo(void *pvParameters);
static void Task1(void *pvParameters);
static void SetupHardware(void);

// Servo values are store as a timer tick.  The timer is set to run at 1250 ticks /millisec
// Servos are pulsed for 0 to 2ms.  You provide the number of "tick" for the pulse length
// this value will need to be scaled depending on your servo.  Typical values are
// 1.0 ms pulse = -180 deg = 1250 ticks
// 1.5 ms pulse =    0 deg = 1875 ticks
// 2.0 ms pulse =  180 deg = 2500 ticks
//   180 degs = 625 ticks
//
// These values are passed to a high priority servo task through a queue.
// The queue holds a 4 byte integer.  The low byte value is the servo number
// The high bytes are the ticks (0 to 2500)
// Thus  queuevalue = ticks<<8 | servoID
// to extract values
//  servoID = queuevalue & 0x000000FF;
//  ticks  =  (queuevalue & 0xFFFFFF00) >>8;
xQueueHandle servoQueue;


// config the servo locations
// servoID is 0 to NSERVOS-1
// configure the pinout locations below - these are arbitrary values for this demo
#define NSERVOS 4
// port for servos
unsigned int ServoPort[NSERVOS] = {IOPORT_G, IOPORT_G, IOPORT_G, IOPORT_G};
// corresponding pins on each port
unsigned int ServoBit[NSERVOS] = {BIT_12, BIT_13, BIT_14, BIT_15};

// Create the  tasks then start the scheduler.

int main(void) {
    int i;
    /* Prepare the hardware to run */
    SetupHardware();

    // set the digital io ports
    // configure the built-in LEDs
    //   LED1 is on RB10
    //	 LED2 is on RB11
    //	 LED3 is on RB12
    //	 LED4 is on RB13
    //Set ports for onboard LEDs to outputs
    PORTSetPinsDigitalOut(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13);
    PORTClearBits(IOPORT_B, BIT_10 | BIT_11 | BIT_12 | BIT_13); // turn off the LEDs

    // set the built-in buttons as inputs
    PORTSetPinsDigitalIn(IOPORT_A, BIT_6 | BIT_7);

    // configure outputs connected to the servos
    for (i = 0; i < NSERVOS; i++) {
        PORTSetPinsDigitalOut(ServoPort[i], ServoBit[i]);
    }

    // queue to send servo info
    servoQueue = xQueueCreate(NSERVOS, 4); // space to update all n servos

    // 10MZ/8 =  1250000 ticks per second
    OpenTimer2(T2_ON | T2_PS_1_8 | T2_SOURCE_INT | T2_IDLE_CON, 0xFFFF);

    // do something in task 1
    xTaskCreate(Task1, // The function that implements the task.
            (signed char *) "Task1", // The text name assigned to the task - for debug only as it is not used by the kernel.
            configMINIMAL_STACK_SIZE + 0, // The size of the stack to allocate to the task.
            // will need to increase this if your task uses much stack space
            (void *) (0), // The parameter passed to the task - just to check the functionality.
            1, // The priority assigned to the task. Currently lower than the timer
            NULL); // The task handle is not required, so NULL is passed.

    // high priority task to handle the servos
    xTaskCreate(Servo, // The function that implements the task.
            (signed char *) "Task2", // The text name assigned to the task - for debug only as it is not used by the kernel.
            configMINIMAL_STACK_SIZE + NSERVOS*4, // The size of the stack to allocate to the task.
            // will need to increase this if your task uses much stack space
            (void *) (0), // The parameter passed to the task - just to check the functionality.
            3, // Set as a high priority task to ensure servo are regularly updated
            NULL); // The task handle is not required, so NULL is passed.


    // Start the tasks and timers running.
    vTaskStartScheduler();

    for (;;);
}

/*-----------------------------------------------------------*/


static void Task1(void *pvParameters) {
    // do something
    // for this demo pressing button 1 increments servo 0
    //  pressing button 2 decrements servo 0
    // this program does not protect against over driving the position of the servos
    // it allows a full range of pulse width, from 0 to 2ms.  Many servos can not handle that full range.

    // servo increment per button press  (50 ticks = 180/625*50= 14.4 degrees)
#define SERVOINC  50  

    // use temp. variable to identify change in button state
    int prevstate = UP;
    int prevstate2 = UP;

    int servo0 = 1875; // typical zero position
    int queueValue;

    while (1) {
        // infinite loop for the task

        // increment servo position
        if (PORTReadBits(IOPORT_A, BIT_6)) {
            // button is now down
            if (prevstate == UP) { // button is now down but was previously UP
                if (servo0 < 2500) { // button has gone from up to down
                    servo0 += SERVOINC;
                    queueValue = (servo0 << 8) | 0x00000000; // servo 0
                    xQueueSend(servoQueue, &queueValue, 0);
                }
            }
            prevstate = DOWN;
        } else {
            prevstate = UP;
        }

        // decrement servo position
        if (PORTReadBits(IOPORT_A, BIT_7)) {
            // button is now down
            if (prevstate2 == UP) { // button has gone from up to down
                if (servo0 > SERVOINC) { // change this to match servo range
                    servo0 -= SERVOINC;
                    queueValue = (servo0 << 8) | 0x00000000; // servo 0
                    xQueueSend(servoQueue, &queueValue, 0);
                }
            }
            prevstate2 = DOWN;
        } else {
            prevstate2 = UP;
        }

        vTaskDelay(50);
    }
}

// high priority task to update the servo outputs
// function run for up to 2ms every SERVOUPDATE ms

static void Servo(void *pvParameters) {

    portTickType pxPreviousWakeTime = xTaskGetTickCount(); // initialize a current time 
    int currTimer; // track timer count
    int queue; // queue value used to update servo command values
    int i;
    int servoMax; // max servo value

    int servoValue[NSERVOS]; // holds current desired servo position
    // servo queue values
    // 0 to 2ms in time tick counts at 1250 ticks per ms
    // typical values are
    // 1.0 ms pulse = -180 deg = 1250 ticks
    // 1.5 ms pulse =    0 deg = 1875 ticks
    // 2.0 ms pulse =  180 deg = 2500 ticks
    // low byte value is the servo number
    // value = ticks<<8 & servoID

    // initialize servos at zero degrees
    for (i = 0; i < NSERVOS; i++) {
        servoValue[i] = 1875;
    }

    // begin main task loop
    // check up update in desired servo position
    // create servo pulse
    // sleep until SERVOUPDATE ms has expired
    while (1) {
        // update servos at a fixed rate
        vTaskDelayUntil(&pxPreviousWakeTime, SERVOUPDATE);

        
        // update desired servo values if available in the queue
        while (xQueueReceive(servoQueue, &queue, 0)) {
            servoValue[queue & 0x000000FF] = (queue & 0xFFFFFF00) >> 8;
        }

        // find the max servo value
        // can sleep task once max value is reached
        servoMax = servoValue[0];
        for (i = 1; i < NSERVOS; i++) {
            if (servoMax < servoValue[i])
                servoMax = servoValue[i];
        }

        // reset the timer
        WriteTimer2(0);

        // Pulse all servos ON
        for (i = 0; i < NSERVOS; i++) {
            PORTSetBits(ServoPort[i], ServoBit[i]);
        }

        // check the timer until the longest servo pulse has expired
        // this burns up to 2ms out of every SERVOUPDATE ms - this corresponds to 20% of CPU time in initialial configuration
        // since this is a high priority task this may affect performance
        do {
            currTimer = ReadTimer2();
            for (i = 0; i < NSERVOS; i++) {
                if (servoValue[i] < currTimer) { // if the servo pulse time has expired, turn it off
                    PORTClearBits(ServoPort[i], ServoBit[i]);
                }
            }
        } while (currTimer <= servoMax); // wait until have handled all of the servos
    }
}

static void SetupHardware(void) {
    /* Set the system and peripheral bus speeds and enable the program cache*/
    SYSTEMConfigPerformance(SYSCLOCK);
    mOSCSetPBDIV(OSC_PB_DIV_8);

    /* Setup to use the external interrupt controller. */
    INTEnableSystemMultiVectoredInt();

    portDISABLE_INTERRUPTS();
}

/*-----------------------------------------------------------
 *  Hooks for RTOS.  Are not required to modify these
 *  May want to turn on LED or some other external signal if an error
 *  occurs
 *-----------------------------------------------------------*/


void vApplicationMallocFailedHook(void) {
    /* vApplicationMallocFailedHook() will only be called if
    configUSE_MALLOC_FAILED_HOOK is set to 1 in FreeRTOSConfig.h.  It is a hook
    function that will get called if a call to pvPortMalloc() fails.
    pvPortMalloc() is called internally by the kernel whenever a task, queue,
    timer or semaphore is created.  It is also called by various parts of the
    demo application.  If heap_1.c or heap_2.c are used, then the size of the
    heap available to pvPortMalloc() is defined by configTOTAL_HEAP_SIZE in
    FreeRTOSConfig.h, and the xPortGetFreeHeapSize() API function can be used
    to query the size of free heap space that remains (although it does not
    provide information on how the remaining heap might be fragmented). */
    taskDISABLE_INTERRUPTS();
    for (;;);
}

/*-----------------------------------------------------------*/

void vApplicationIdleHook(void) {
    /* vApplicationIdleHook() will only be called if configUSE_IDLE_HOOK is set
    to 1 in FreeRTOSConfig.h.  It will be called on each iteration of the idle
    task.  It is essential that code added to this hook function never attempts
    to block in any way (for example, call xQueueReceive() with a block time
    specified, or call vTaskDelay()).  If the application makes use of the
    vTaskDelete() API function (as this demo application does) then it is also
    important that vApplicationIdleHook() is permitted to return to its calling
    function, because it is the responsibility of the idle task to clean up
    memory allocated by the kernel to any task that has since been deleted. */
}

/*-----------------------------------------------------------*/

void vApplicationStackOverflowHook(xTaskHandle pxTask, signed char *pcTaskName) {
    (void) pcTaskName;
    (void) pxTask;

    /* Run time task stack overflow checking is performed if
    configCHECK_FOR_STACK_OVERFLOW is defined to 1 or 2.  This hook	function is
    called if a task stack overflow is detected.  Note the system/interrupt
    stack is not checked. */
    taskDISABLE_INTERRUPTS();
    for (;;);
}

/*-----------------------------------------------------------*/

void vApplicationTickHook(void) {
    /* This function will be called by each tick interrupt if
    configUSE_TICK_HOOK is set to 1 in FreeRTOSConfig.h.  User code can be
    added here, but the tick hook is called from an interrupt context, so
    code must not attempt to block, and only the interrupt safe FreeRTOS API
    functions can be used (those that end in FromISR()). */
}

/*-----------------------------------------------------------*/

void _general_exception_handler(unsigned long ulCause, unsigned long ulStatus) {
    /* This overrides the definition provided by the kernel.  Other exceptions
    should be handled here. */
    for (;;);
}
/*-----------------------------------------------------------*/

